import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:welltested/cli/ast_visitor.dart';
import 'package:welltested/generators/saver.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';

import '../generators/unit_generator.dart';
import '../src/repositories/dart_server_repository.dart';
import '../src/utilities/utilities.dart';
import '../src/visitors/class_visitor.dart';

class CLIOperation {
  static Future<void> saveUnit({required PubspecData pubspecData}) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = Directory(PathUtils.libPath)
        .listSync(recursive: true)
        .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );

    List<Future> saveFutures = [];
    final testRunInfo = TestRunInfo.newRun();
    for (var file in fileList) {
      saveFutures.add(
        runSaveUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo),
      );
    }
    wtLog.startSpinner("Saving Unit Tests");
    await Future.wait(saveFutures);
    wtLog.stopSpinner(
        message:
            '✅ Succeeded. ${testRunInfo.updated} were saved.\nNote: Saving working tests helps improve future test generations. Keep it up!');
    dartServerRepository.dispose();
  }

  static Future<void> generateUnit({
    required PubspecData pubspecData,
    String? filePath,
    String? className,
    String? methodName,
  }) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = filePath != null
        ? [File(filePath)]
        : (Directory(PathUtils.libPath).listSync(recursive: true))
            .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );
    List<Future> generationFutures = [];
    final testRunInfo = TestRunInfo.newRun();

    for (var file in fileList) {
      generationFutures.add(
        runUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo,
            className: className, methodName: methodName),
      );
    }
    wtLog.startSpinner('Generating unit tests');
    await Future.wait(generationFutures);

    wtLog.stopSpinner(
        message:
            '✅ Succeeded. Please find the generated tests in `/test` folder.\nCreated: ${testRunInfo.created}\nUpdated:${testRunInfo.updated}\nFailed:${testRunInfo.failed}');

    dartServerRepository.dispose();
    await runMockitoCommand();
  }
}

Future runUnit(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
    {String? className,
    String? methodName}) async {
  final astVisitor = WelltestedASTVisitor();
  final resolvedUnit = await file.resolveFile(collection);
  CompilationUnit unit = resolvedUnit.unit;
  unit.visitChildren(astVisitor);

  List<Future> generationFutures = [];

  for (final declaration in unit.declarations) {
    if (declaration is ClassDeclaration) {
      if (className != null && declaration.name.toString() != className)
        continue;
      final element = declaration.declaredElement!;
      if (!isAnnotationPresent(element)) continue;
      generationFutures.add(UnitGenerator().generateForAnnotatedElement(
        element: element,
        annotation: element.metadata.firstWhere(
            (element) => element.element!.displayName == "Welltested"),
        unit: unit,
        pubspecData: pubspecData,
        dartServerRepository: dartServerRepository,
        classDeclaration: declaration,
        testRunInfo: testRunInfo,
        classFile: file as File,
        requestedFunction: methodName,
      ));
    }
  }
  await Future.wait(generationFutures);
}

Future runSaveUnit(
  FileSystemEntity file,
  AnalysisContextCollection collection,
  PubspecData pubspecData,
  DartServerRepository dartServerRepository,
  TestRunInfo testRunInfo,
) async {
  final astVisitor = WelltestedASTVisitor();
  final resolvedUnit = await file.resolveFile(collection);
  CompilationUnit unit = resolvedUnit.unit;
  unit.visitChildren(astVisitor);

  List<Future> generationFutures = [];

  for (final declaration in unit.declarations.whereType<ClassDeclaration>()) {
    final element = declaration.declaredElement!;
    if (!isAnnotationPresent(element)) continue;
    generationFutures.add(WelltestedSaver().saveForAnnotatedElement(
      element: element,
      annotation: element.metadata.firstWhere(
          (element) => element.element!.displayName == "Welltested"),
      unit: unit,
      pubspecData: pubspecData,
      dartServerRepository: dartServerRepository,
      classDeclaration: declaration,
      testRunInfo: testRunInfo,
      classFile: file as File,
    ));

    continue;
  }
  await Future.wait(generationFutures);
}

runMockitoCommand() async {
  wtLog.startSpinner("Generating mocks");
  await Future.delayed(Duration(seconds: 1));
  final process = await Process.run(
      "dart", ["run", "build_runner", "build", "--delete-conflicting-outputs"],
      runInShell: true);

  if (process.exitCode != 0) {
    wtLog.stopSpinner(
        message: "Failed to generate mocks:",
        severity: MessageSeverity.Warning);
    wtLog.log(process.stderr);
    wtLog.log(
        'Note: You may also directly use the command `dart run build_runner build` to generate the mocks after fixing the above errors.');
  } else {
    wtLog.stopSpinner(
        message: 'Successfully generated mocks.',
        severity: MessageSeverity.Info);
  }
}

Future<Map<String, MethodDeclaration>> generateForElement(
    Element element) async {
  final classVisitor = ClassVisitor();
  final Map<String, MethodDeclaration> methodAstNodes = {};

  element.visitChildren(classVisitor);
  for (final method in classVisitor.methods.entries) {
    final node =
        (await getResolvedAstNodeForElement(method.value)) as MethodDeclaration;
    methodAstNodes[method.key] = node;
  }
  return methodAstNodes;
}

Future<AstNode> getResolvedAstNodeForElement(
  Element element,
) async {
  final resolvedLib =
      await element.session!.getResolvedLibraryByElement(element.library!);
  final astNode = (resolvedLib as ResolvedLibraryResult)
      .getElementDeclaration(element)!
      .node;
  return astNode;
}
