/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/src/executors/ast_elm_executor.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../src/visitors/class_visitor.dart';

class WelltestedSaver {
  Future<void> saveForAnnotatedElement({
    required ClassElement element,
    required ElementAnnotation annotation,
    required CompilationUnit unit,
    required PubspecData pubspecData,
    required DartServerRepository dartServerRepository,
    required ClassDeclaration classDeclaration,
    required TestRunInfo testRunInfo,
    required File classFile,
  }) async {
    final generationRepository = GenerationRepository();

    ///Executors
    final fixerExecutor = FixerExecutor(dartServerRepository
        .server); //TODO: Create methods and don't pass server directly
    final astElementExecutor = AstElementExecutor();
    final unitTestExecutor = UnitTestExecutor(generationRepository);
    final cacheManager = CacheManager();

    final List<String> excludedMethods = annotation
        .computeConstantValue()!
        .getField("excludedMethods")!
        .toListValue()!
        .map((e) => e.toStringValue()!)
        .toList();

    final classAstNode = classDeclaration;
    final fileAstNode = classAstNode.parent;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final classVisitor = ClassVisitor();
    element.visitChildren(classVisitor);
    final methodAstNodes = await astElementExecutor.methodNodesFromElement(
        element); //TODO: probably an extension is better fit
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      try {
        final method = methodAstNodes.entries.elementAt(i);

        final testRequestData = await unitTestExecutor.prepareRequest(
          method: method,
          fileAstNode: fileAstNode,
          classAstNode: classAstNode,
          fileImport: fileImport,
          className: classVisitor.className,
          methodAstNodes: methodAstNodes,
          excludedMethods: excludedMethods,
          unit: unit,
        );
        if (testRequestData == null) continue;

        final file =
            PathUtils.correspondingPermanentTestFile(method.key, classFile);
        if (!(await file.exists())) continue;
        final testIsSaved = await cacheManager.checkIfTestIsSaved(
            testRequestData.codeInScrutity(excludeAnnotations: false),
            classFile.path,
            testRequestData.functionName,
            className: testRequestData.className);
        if (testIsSaved) continue; //this test is already saved.
        var hasErrors =
            await fixerExecutor.hasCompilationErrors(file.absolute.path);

        if (hasErrors) {
          wtLog.log(
              "⚠️ File ${PathUtils.relatetiveLibFilePath(file.path)} has errors and will not be updated");
          continue;
        }
        final fixedTest = await file.readAsString();
        await unitTestExecutor.updateTest(testRequestData, fixedTest);
        await cacheManager.cacheTest(
            code: testRequestData.codeInScrutity(excludeAnnotations: false),
            codeFilePath: classFile.path,
            test: fixedTest,
            testFilePath: file.path,
            functionName: testRequestData.functionName,
            className: testRequestData.className,
            saved: true);
        testRunInfo.incrementFromAction(TestGenerationStatus.update);
      } catch (e, stackTrace) {
        wtTelemetry.trackError(
            severity: Severity.error, error: e, stackTrace: stackTrace);
        wtLog.warning(
            "Error saving unit tests for method: ${methodAstNodes.entries.elementAt(i).key}");
        wtLog.log(e.toString());
      }
    }
  }
}
