/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */
import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:source_gen/source_gen.dart';
import 'package:welltested/src/executors/fixer_executor.dart';
import 'package:welltested/src/executors/unit_test_executor.dart';
import 'package:welltested/src/executors/writer_executor.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/repositories/dart_server_repository.dart';
import 'package:welltested/src/repositories/generations_repository.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/repositories/writer_repository.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import 'package:welltested/src/visitors/class_visitor.dart';

import '../src/executors/ast_elm_executor.dart';

class UnitGenerator {
  bool initialized = false;
  Future<FutureOr<String>> generateForAnnotatedElement(
      {required Element element,
      required ElementAnnotation annotation,
      required CompilationUnit unit,
      required ClassDeclaration classDeclaration,
      required PubspecData pubspecData,
      required DartServerRepository dartServerRepository,
      required File classFile,
      required String? requestedFunction,
      required TestRunInfo testRunInfo}) async {
    if (element is! ClassElement) {
      final String friendlyName = element.displayName;
      throw InvalidGenerationSourceError(
        'Generator cannot target `$friendlyName`.',
        todo: 'Remove the [Welltested] annotation from `$friendlyName`.',
      );
    }

    final generationRepository = GenerationRepository();
    final writerRepository = WriterRepository();

    ///Executors
    final fixerExecutor = FixerExecutor(dartServerRepository
        .server); //TODO: Create methods and don't pass server directly
    final astElementExecutor = AstElementExecutor();
    final unitTestExecutor = UnitTestExecutor(generationRepository);
    final cacheManager = CacheManager();

    final writerExecutor = WriterExecutor(
        writerRepository, dartServerRepository, fixerExecutor, cacheManager);
    final List<String> excludedMethods = annotation
        .computeConstantValue()!
        .getField("excludedMethods")!
        .toListValue()!
        .map((e) => e.toStringValue()!)
        .toList();

    final classAstNode = classDeclaration;
    final fileAstNode = classAstNode.parent;
    final fileImport =
        'package:${pubspecData.packageName}/${PathUtils.relatetiveLibFilePath(classFile.path)}';

    final classVisitor = ClassVisitor();
    element.visitChildren(classVisitor);
    final methodAstNodes = await astElementExecutor.methodNodesFromElement(
        element); //TODO: probably an extension is better fit
    for (int i = 0; i < methodAstNodes.entries.length; i++) {
      try {
        final method = methodAstNodes.entries.elementAt(i);
        if (requestedFunction != null && method.key != requestedFunction)
          continue;

        final testRequestData = await unitTestExecutor.prepareRequest(
          method: method,
          fileAstNode: fileAstNode,
          classAstNode: classAstNode,
          fileImport: fileImport,
          className: classVisitor.className,
          methodAstNodes: methodAstNodes,
          excludedMethods: excludedMethods,
          unit: unit,
        );
        if (testRequestData == null) continue;
        wtLog.updateSpinnerMessage(
            "Generating Test for ${testRequestData.functionName}");
        final file =
            PathUtils.correspondingPermanentTestFile(method.key, classFile);

        final existStatus = await cacheManager.testExistStatus(
            testRequestData.codeInScrutity(excludeAnnotations: false),
            classFile.path,
            testRequestData.functionName,
            className: testRequestData.className);
        bool update = false;
        late TestGenerationStatus testGenerationStatus;
        switch (existStatus) {
          case TestExistStatus.generationExists:
            testGenerationStatus = TestGenerationStatus.omit;
            continue;
          case TestExistStatus.onlyRequestExists:
            testRequestData.forceGenerate = true;
            testGenerationStatus = TestGenerationStatus.forceUpdate;
            break;
          case TestExistStatus.outdatedCodeExists:
            update = true;
            testGenerationStatus = TestGenerationStatus.update;
            break;
          case TestExistStatus.doesNotExist:
            testGenerationStatus = TestGenerationStatus.create;
            break;
        }

        final unitTests =
            await unitTestExecutor.buildUnitTests(testRequestData);
        String? completion = unitTests['fixed_completion'];
        late File testFile;
        if (completion != null) {
          testFile =
              await writerExecutor.outputGeneration(file.path, completion);
        } else {
          completion = unitTests['model_completion'];
          if (completion == null) {
            wtLog.warning(
                "⛔️ Sorry, we can't process ${testRequestData.className} at the moment. Please try again Later.");
            continue;
          }
          testFile = await writerExecutor.processGeneration(
              file.path, completion, method.key);
        }
        await cacheManager.cacheTest(
            code: testRequestData.codeInScrutity(excludeAnnotations: false),
            codeFilePath: classFile.path,
            test: completion,
            testFilePath: testFile.path,
            className: testRequestData.className,
            functionName: testRequestData.functionName);
        testRunInfo.incrementFromAction(testGenerationStatus);
        wtLog.log(
            '✔️ Tests ${update ? 'updated' : 'generated'} for ${method.key}');
      } catch (e, stackTrace) {
        testRunInfo.incrementFromAction(TestGenerationStatus.failed);
        wtTelemetry.trackError(
            severity: Severity.error, error: e, stackTrace: stackTrace);
        wtLog.warning(
            "Error generating unit tests for method: ${methodAstNodes.entries.elementAt(i).key}");
        wtLog.log(e.toString());
      }
    }

    return '';
  }
}
