/*
 * Created on May 1st 2023
 *
 * Copyright (C) 2023 Welltested AI - All Rights Reserved
 * The code in this file is Intellectutal Property of Welltested AI.
 * and can't be edited, redistributed or used without a valid license of it's product (Welltested)[https://welltested.ai]
 */

import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import '../utilities/typedef.dart';

extension ClassElementExt on ClassElement {
  bool get isSimple {
    return !this.isPrivate &&
        !this.hasProtected &&

        ///TODO: Replace [!this.isAbstract] with [this.isConstructable] after analyzer 5.7.1
        !this.isAbstract &&
        this.constructors.indexWhere((e) => !e.isFactory) !=
            -1; //Has non factory constructors
  }

  ///Contains elements that aren't part of core data type
  List<ClassElement> get nonDartElements {
    List<ClassElement> nonSDK = [];
    this.fields.forEach((field) {
      final fieldElement = field.type.element;
      if (fieldElement is ClassElement) {
        if (!fieldElement.library.isInSdk) {
          nonSDK.add(fieldElement);
        }
      }
    });
    return nonSDK;
  }

  ///Returns the source code of the class without factory constructors and methods.
  Future<String> simplefiedSourceFromUnit(CompilationUnit unit) async {
    String classCode = unit.toSource();
    this.source.toString();
    final constructors = unit.declarations.whereType<ConstructorDeclaration>();
    for (final constructor in constructors) {
      if (constructor.factoryKeyword == null) {
        final replaceString = constructor.toSource();
        classCode = classCode.replaceAll(replaceString, '');
      }
    }

    final methods = unit.declarations.whereType<MethodDeclaration>();
    for (final method in methods) {
      final replaceString = method.toSource();
      classCode = classCode.replaceAll(replaceString, '');
    }

    return classCode;
  }
}

extension ClassDeclarationExtension on ClassDeclaration {
  List<AstNode>? checkIfWidget() {
    if (this.extendsClause?.superclass.name == 'StatelessWidget' ||
        this.extendsClause?.superclass.name.name == 'ConsumerStatefulWidget' ||
        this.extendsClause?.superclass.name.name == 'StatefulWidget' ||
        (this.extendsClause?.superclass.name.name.contains('GetWidget') ??
            false)) {
      return parent?.childEntities
          .where((element) {
            return element is ImportDirective ||
                element == this ||
                (element is ClassDeclaration &&
                    (element.extendsClause?.superclass.name.name
                            .contains('State') ??

                        ///TODO: Consider widgets with other extending classes
                        false));
          })
          .map((e) => e as AstNode)
          .toList();
    } else
      return null;
  }
}

extension Processing on List<ClassElement> {
  Future<List> processModels(CompilationUnit unit) async {
    final List<ClassElement> complexModels = [];
    final List<ClassElement> simpleModels = [];

    for (final element in this) {
      if (!element.isSimple) {
        complexModels.add(element);
        continue;
      }
      final nonSDKElements = element.nonDartElements;
      if (nonSDKElements.isEmpty) {
        simpleModels.add(element);
        continue;
      }

      bool isComplexModel = false;
      for (final nonSDKElement in nonSDKElements) {
        final moreNonSDKElements = nonSDKElement.nonDartElements;
        if (moreNonSDKElements.isNotEmpty) {
          isComplexModel = true;
          break;
        }
      }
      if (isComplexModel) {
        complexModels.add(element);
      } else {
        //TODO: Maybe define a max field limit length of comined models
        simpleModels.add(element);
        simpleModels.addAll(nonSDKElements);
      }
    }
    final List<String> simpleModelCode = [];

    for (final simpleModel in simpleModels) {
      simpleModelCode.add(await simpleModel.simplefiedSourceFromUnit(unit));
    }

    return [simpleModelCode, complexModels];
  }
}

extension IntegrationsTestExtensions on ClassElement {
  Future<String> setup(ElementToNode elementToNode) async {
    final descriptionElm = fields.where((field) => field.name == 'setup').first;
    final descriptionNode = ((await elementToNode(
      descriptionElm.getter!.nonSynthetic,
    ) as MethodDeclaration)
            .body as ExpressionFunctionBody)
        .expression;
    final description = (descriptionNode as SimpleStringLiteral).value;
    return description;
  }

  Future<List<String>> testcases(ElementToNode elementToNode) async {
    final testCasesElm =
        fields.where((field) => field.name == 'testcases').first;
    final testCasesNode =
        ((await elementToNode(testCasesElm.getter!.nonSynthetic)
                    as MethodDeclaration)
                .body as ExpressionFunctionBody)
            .expression;
    final testCases = (testCasesNode as ListLiteral)
        .elements
        .map((e) => (e as SimpleStringLiteral).value)
        .toList();
    return testCases;
  }
}
